@@ -17,16 +17,25 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 # USA.
-from PyQt5.QtCore import QObject
-from picard import config
-from picard.plugin import PluginFunctions, PluginPriority
-from picard.similarity import similarity2
-from picard.util import (
-    linear_combination_of_weights,
+from collections import namedtuple
+from collections.abc import (
+    Iterable,
+    MutableMapping,
 )
-from picard.util.tags import PRESERVED_TAGS
+
+from PyQt5.QtCore import QObject
+
+from picard import config
 from picard.mbjson import artist_credit_from_node
+from picard.plugin import (
+    PluginFunctions,
+    PluginPriority,
+)
+from picard.similarity import similarity2
+from picard.util import linear_combination_of_weights
 from picard.util.imagelist import ImageList
+from picard.util.tags import PRESERVED_TAGS
+
 
 MULTI_VALUED_JOINER = '; '
 
@@ -42,8 +51,11 @@ MULTI_VALUED_JOINER = '; '
 # 20000 50000 0.0
 LENGTH_SCORE_THRES_MS = 30000
 
+SimMatchTrack = namedtuple('SimMatchTrack', 'similarity releasegroup release track')
+SimMatchRelease = namedtuple('SimMatchRelease', 'similarity release')
 
-class Metadata(dict):
+
+class Metadata(MutableMapping):
 
     """List of metadata items with dict-like access."""
 
@@ -57,49 +69,35 @@ class Metadata(dict):
 
     multi_valued_joiner = MULTI_VALUED_JOINER
 
-    def __init__(self):
-        super().__init__()
-        self.images = ImageList()
+    def __init__(self, *args, deleted_tags=None, images=None, length=None, **kwargs):
+        self._store = dict()
         self.deleted_tags = set()
         self.length = 0
+        self.images = ImageList()
+        self.has_common_images = True
+
+        d = dict(*args, **kwargs)
+        for k, v in d.items():
+            self[k] = v
+        if images is not None:
+            for image in images:
+                self.images.append(image)
+        if deleted_tags is not None:
+            for tag in deleted_tags:
+                del self[tag]
+        if length is not None:
+            self.length = int(length)
 
     def __bool__(self):
-        return bool(len(self) or len(self.images))
+        return bool(len(self))
 
-    def append_image(self, coverartimage):
-        self.images.append(coverartimage)
-
-    def set_front_image(self, coverartimage):
-        # First remove all front images
-        self.images[:] = [img for img in self.images if not img.is_front_image()]
-        self.images.append(coverartimage)
-
-    @property
-    def images_to_be_saved_to_tags(self):
-        if not config.setting["save_images_to_tags"]:
-            return ()
-        images = [img for img in self.images if img.can_be_saved_to_tags]
-        if config.setting["embed_only_one_front_image"]:
-            front_image = self.get_single_front_image(images)
-            if front_image:
-                return front_image
-        return images
-
-    def get_single_front_image(self, images=None):
-        if not images:
-            images = self.images
-        for img in images:
-            if img.is_front_image():
-                return [img]
-        return []
-
-    def remove_image(self, index):
-        self.images.pop(index)
+    def __len__(self):
+        return len(self._store) + len(self.images)
 
     @staticmethod
     def length_score(a, b):
-        return (1.0 - min(abs(a - b), LENGTH_SCORE_THRES_MS) /
-                float(LENGTH_SCORE_THRES_MS))
+        return (1.0 - min(abs(a - b),
+                LENGTH_SCORE_THRES_MS) / float(LENGTH_SCORE_THRES_MS))
 
     def compare(self, other):
         parts = []
@@ -123,7 +121,9 @@ class Metadata(dict):
                 else:
                     score = similarity2(a, b)
                 parts.append((score, weight))
-
+            elif (a and name in other.deleted_tags
+                  or b and name in self.deleted_tags):
+                parts.append((0, weight))
         return linear_combination_of_weights(parts)
 
     def compare_to_release(self, release, weights):
@@ -132,7 +132,10 @@ class Metadata(dict):
         linear combination of weights that the metadata matches a certain album.
         """
         parts = self.compare_to_release_parts(release, weights)
-        return (linear_combination_of_weights(parts), release)
+        sim = linear_combination_of_weights(parts)
+        if 'score' in release:
+            sim *= release['score'] / 100
+        return SimMatchRelease(similarity=sim, release=release)
 
     def compare_to_release_parts(self, release, weights):
         parts = []
@@ -147,18 +150,11 @@ class Metadata(dict):
 
         try:
             a = int(self["totaltracks"])
-        except (ValueError, KeyError):
-            pass
-        else:
-            try:
-                if "title" in weights:
-                    b = release['media'][0]['track-count']
-                else:
-                    b = release['track-count']
-            except KeyError:
-                b = 0
+            b = release['track-count']
             score = 0.0 if a > b else 0.3 if a < b else 1.0
             parts.append((score, weights["totaltracks"]))
+        except (ValueError, KeyError):
+            pass
 
         preferred_countries = config.setting["preferred_release_countries"]
         preferred_formats = config.setting["preferred_release_formats"]
@@ -220,12 +216,13 @@ class Metadata(dict):
 
         if 'title' in self:
             a = self['title']
-            b = track['title']
+            b = track.get('title', '')
             parts.append((similarity2(a, b), weights["title"]))
 
         if 'artist' in self:
             a = self['artist']
-            b = artist_credit_from_node(track['artist-credit'])[0]
+            artist_credits = track.get('artist-credit', [])
+            b = artist_credit_from_node(artist_credits)[0]
             parts.append((similarity2(a, b), weights["artist"]))
 
         a = self.length
@@ -240,85 +237,120 @@ class Metadata(dict):
 
         if not releases:
             sim = linear_combination_of_weights(parts)
-            return (sim, None, None, track)
-
-        result = (-1,)
+            if 'score' in track:
+                sim *= track['score'] / 100
+            return SimMatchTrack(similarity=sim, releasegroup=None, release=None, track=track)
 
+        result = SimMatchTrack(similarity=-1, releasegroup=None, release=None, track=None)
         for release in releases:
             release_parts = self.compare_to_release_parts(release, weights)
             sim = linear_combination_of_weights(parts + release_parts)
-            if sim > result[0]:
+            if 'score' in track:
+                sim *= track['score'] / 100
+            if sim > result.similarity:
                 rg = release['release-group'] if "release-group" in release else None
-                result = (sim, rg, release, track)
+                result = SimMatchTrack(similarity=sim, releasegroup=rg, release=release, track=track)
         return result
 
-    def copy(self, other):
+    def copy(self, other, copy_images=True):
         self.clear()
-        self.update(other)
+        self._update_from_metadata(other, copy_images)
 
-    def update(self, other):
-        for key in other.keys():
-            self.set(key, other.getall(key)[:])
-        if other.images:
-            self.images = other.images[:]
+    def update(self, *args, **kwargs):
+        one_arg = len(args) == 1
+        if one_arg and isinstance(args[0], self.__class__):
+            self._update_from_metadata(args[0])
+        elif one_arg and isinstance(args[0], MutableMapping):
+            # update from MutableMapping (ie. dict)
+            for k, v in args[0].items():
+                self[k] = v
+        elif args or kwargs:
+            # update from a dict-like constructor parameters
+            for k, v in dict(*args, **kwargs).items():
+                self[k] = v
+        else:
+            # no argument, raise TypeError to mimic dict.update()
+            raise TypeError("descriptor 'update' of '%s' object needs an argument" % self.__class__.__name__)
+
+    def _update_from_metadata(self, other, copy_images=True):
+        for k, v in other.rawitems():
+            self.set(k, v[:])
+
+        for tag in other.deleted_tags:
+            del self[tag]
+
+        if copy_images and other.images:
+            self.images = other.images.copy()
         if other.length:
             self.length = other.length
 
-        self.deleted_tags.update(other.deleted_tags)
-        # Remove deleted tags from UI on save
-        for tag in other.deleted_tags:
-            self.pop(tag, None)
-
     def clear(self):
-        super().clear()
+        self._store.clear()
         self.images = ImageList()
         self.length = 0
+        self.clear_deleted()
+
+    def clear_deleted(self):
         self.deleted_tags = set()
 
     def getall(self, name):
-        return super().get(name, [])
+        return self._store.get(name, [])
 
-    def get(self, name, default=None):
-        values = super().get(name, None)
+    def getraw(self, name):
+        return self._store[name]
+
+    def get(self, key, default=None):
+        values = self._store.get(key, None)
         if values:
             return self.multi_valued_joiner.join(values)
         else:
             return default
 
+    def __contains__(self, name):
+        return self._store.__contains__(name)
+
     def __getitem__(self, name):
         return self.get(name, '')
 
     def set(self, name, values):
-        super().__setitem__(name, values)
-        if name in self.deleted_tags:
-            self.deleted_tags.remove(name)
+        self._store[name] = values
+        self.deleted_tags.discard(name)
 
     def __setitem__(self, name, values):
-        if not isinstance(values, list):
+        if isinstance(values, str) or not isinstance(values, Iterable):
             values = [values]
-        values = [string_(value) for value in values if value]
-        if len(values):
+        values = [str(value) for value in values if value or value == 0]
+        if values:
             self.set(name, values)
-        else:
-            self.delete(name)
+        elif name in self._store:
+            del self[name]
+
+    def __delitem__(self, name):
+        try:
+            del self._store[name]
+        except KeyError:
+            pass
+        finally:
+            self.deleted_tags.add(name)
 
     def add(self, name, value):
         if value or value == 0:
-            self.setdefault(name, []).append(value)
-            if name in self.deleted_tags:
-                self.deleted_tags.remove(name)
+            self._store.setdefault(name, []).append(value)
+            self.deleted_tags.discard(name)
 
     def add_unique(self, name, value):
         if value not in self.getall(name):
             self.add(name, value)
 
     def delete(self, name):
-        if name in self:
-            self.pop(name, None)
-        self.deleted_tags.add(name)
+        """Deprecated: use del directly"""
+        del self[name]
+
+    def __iter__(self):
+        return iter(self._store)
 
     def items(self):
-        for name, values in super().items():
+        for name, values in self._store.items():
             for value in values:
                 yield name, value
 
@@ -328,12 +360,12 @@ class Metadata(dict):
         >>> m.rawitems()
         [("key1", ["value1", "value2"]), ("key2", ["value3"])]
         """
-        return dict.items(self)
+        return self._store.items()
 
     def apply_func(self, func):
-        for key, values in self.rawitems():
-            if key not in PRESERVED_TAGS:
-                self[key] = [func(value) for value in values]
+        for name, values in list(self.rawitems()):
+            if name not in PRESERVED_TAGS:
+                self[name] = [func(value) for value in values]
 
     def strip_whitespace(self):
         """Strip leading/trailing whitespace.
@@ -346,11 +378,17 @@ class Metadata(dict):
         >>> m["foo"]
         "bar"
         """
-        self.apply_func(lambda s: s.strip())
+        self.apply_func(str.strip)
+
+    def __repr__(self):
+        return "%s(%r, deleted_tags=%r, length=%r, images=%r)" % (self.__class__.__name__, self._store, self.deleted_tags, self.length, self.images)
+
+    def __str__(self):
+        return ("store: %r\ndeleted: %r\nimages: %r\nlength: %r" % (self._store, self.deleted_tags, [str(img) for img in self.images], self.length))
 
 
-_album_metadata_processors = PluginFunctions()
-_track_metadata_processors = PluginFunctions()
+_album_metadata_processors = PluginFunctions(label='album_metadata_processors')
+_track_metadata_processors = PluginFunctions(label='track_metadata_processors')
 
 
 def register_album_metadata_processor(function, priority=PluginPriority.NORMAL):
@@ -367,5 +405,5 @@ def run_album_metadata_processors(album_object, metadata, release):
     _album_metadata_processors.run(album_object, metadata, release)
 
 
-def run_track_metadata_processors(track_object, metadata, release, track):
-    _track_metadata_processors.run(track_object, metadata, track, release)
+def run_track_metadata_processors(album_object, metadata, track, release=None):
+    _track_metadata_processors.run(album_object, metadata, track, release)
