@@ -19,6 +19,7 @@
 
 from functools import partial
 import os
+import re
 
 from PyQt5 import (
     QtCore,
@@ -66,7 +67,12 @@ class ActiveLabel(QtWidgets.QLabel):
         if self.active and event.button() == QtCore.Qt.LeftButton:
             self.clicked.emit()
 
-    def dragEnterEvent(self, event):
+    @staticmethod
+    def dragEnterEvent(event):
+        event.acceptProposedAction()
+
+    @staticmethod
+    def dragMoveEvent(event):
         event.acceptProposedAction()
 
     def dropEvent(self, event):
@@ -80,19 +86,12 @@ class ActiveLabel(QtWidgets.QLabel):
         if not dropped_data:
             dropped_data = bytes(mime_data.data('application/x-qt-image'))
 
-        if dropped_data:
-            try:
-                mime = imageinfo.identify(dropped_data)[2]
-                if mime in ('image/jpeg', 'image/png'):
-                    accepted = True
-                    log.debug("Dropped %s mime data (%d bytes)", mime,
-                              len(dropped_data or ''))
-                    self.image_dropped.emit(QtCore.QUrl(''), dropped_data)
-            except imageinfo.IdentificationError as e:
-                log.debug("Image identification failed: %s", e)
+        if not dropped_data:
+            # Maybe we can get something useful from a dropped HTML snippet.
+            dropped_data = bytes(mime_data.data('text/html'))
 
         if not accepted:
-            for url in event.mimeData().urls():
+            for url in mime_data.urls():
                 if url.scheme() in ('https', 'http', 'file'):
                     accepted = True
                     log.debug("Dropped %s url (with %d bytes of data)",
@@ -280,11 +279,15 @@ class CoverArtThumbnail(ActiveLabel):
 
 
 def set_image_replace(obj, coverartimage):
-    obj.metadata.set_front_image(coverartimage)
+    obj.metadata.images.strip_front_images()
+    obj.metadata.images.append(coverartimage)
 
 
 def set_image_append(obj, coverartimage):
-    obj.metadata.append_image(coverartimage)
+    obj.metadata.images.append(coverartimage)
+
+
+HTML_IMG_SRC_REGEX = re.compile(r'<img .*?src="(.*?)"', re.UNICODE)
 
 
 class CoverArtBox(QtWidgets.QGroupBox):
@@ -391,26 +394,57 @@ class CoverArtBox(QtWidgets.QGroupBox):
                 self.load_remote_image(url, mime, data)
 
     def on_remote_image_fetched(self, url, data, reply, error, fallback_data=None):
+        if error:
+            log.error("Failed loading remote image from %s: %s", url, reply.errorString())
+            if fallback_data:
+                self._load_fallback_data(url, fallback_data)
+            return
+
         data = bytes(data)
         mime = reply.header(QtNetwork.QNetworkRequest.ContentTypeHeader)
+        # Some sites return a mime type with encoding like "image/jpeg; charset=UTF-8"
+        mime = mime.split(';')[0]
         url_query = QtCore.QUrlQuery(url.query())
+        # If mime indicates only binary data we can try to guess the real mime type
+        if mime in ('application/octet-stream', 'binary/data'):
+            mime = imageinfo.identify(data)[2]
         if mime in ('image/jpeg', 'image/png'):
             self.load_remote_image(url, mime, data)
         elif url_query.hasQueryItem("imgurl"):
             # This may be a google images result, try to get the URL which is encoded in the query
             url = QtCore.QUrl(url_query.queryItemValue("imgurl", QtCore.QUrl.FullyDecoded))
             self.fetch_remote_image(url)
+        elif url_query.hasQueryItem("mediaurl"):
+            # Bing uses mediaurl
+            url = QtCore.QUrl(url_query.queryItemValue("mediaurl", QtCore.QUrl.FullyDecoded))
+            self.fetch_remote_image(url)
         else:
             log.warning("Can't load remote image with MIME-Type %s", mime)
             if fallback_data:
-                # Tests for image format obtained from file-magic
-                try:
-                    mime = imageinfo.identify(fallback_data)[2]
-                except imageinfo.IdentificationError as e:
-                    log.error("Unable to identify dropped data format: %s" % e)
-                else:
-                    self.load_remote_image(url, mime, fallback_data)
-                    log.debug("Trying the dropped %s data", mime)
+                self._load_fallback_data(url, fallback_data)
+
+    def _load_fallback_data(self, url, data):
+        # Tests for image format obtained from file-magic
+        try:
+            mime = imageinfo.identify(data)[2]
+        except imageinfo.IdentificationError as e:
+            log.warning("Unable to identify dropped data format: %s" % e)
+        else:
+            log.debug("Trying the dropped %s data", mime)
+            self.load_remote_image(url, mime, data)
+            return
+
+        # Try getting image out of HTML (e.g. for Goole image search detail view)
+        try:
+            html = data.decode()
+            match = re.search(HTML_IMG_SRC_REGEX, html)
+            if match:
+                url = QtCore.QUrl(match.group(1))
+        except UnicodeDecodeError as e:
+            log.warning("Unable to decode dropped data format: %s" % e)
+        else:
+            log.debug("Trying URL parsed from HTML: %s", url.toString())
+            self.fetch_remote_image(url)
 
     def load_remote_image(self, url, mime, data):
         try:
