@@ -17,21 +17,34 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
-import webbrowser
-from PyQt5 import QtWidgets
-from picard.const import PICARD_URLS
+"""A webbrowser extension for Picard.
 
+It handles and displays errors in PyQt and also adds a utility function for opening Picard URLS.
 """
-A webbrowser extension for Picard. It handles and displays errors
-in PyQt and also adds a utility function for opening Picard URLS.
-"""
+
+from sys import version_info
+import webbrowser
+
+from PyQt5 import QtWidgets
+
+from picard import log
+from picard.const import PICARD_URLS
 
 
 def open(url):
     try:
-        webbrowser.open(string_(url))
+        webbrowser.open(url)
     except webbrowser.Error as e:
         QtWidgets.QMessageBox.critical(None, _("Web Browser Error"), _("Error while launching a web browser:\n\n%s") % (e,))
+    except TypeError:
+        if version_info.major == 3 and version_info.minor == 7 and version_info.micro == 0:
+            # See https://bugs.python.org/issue31014, webbrowser.open doesn't
+            # work on 3.7.0 the first time it's called. The initialization code
+            # in it will be skipped after the first call, making it possibly to
+            # use it, although it might not accurately identify the users
+            # preferred browser.
+            log.info("Working around https://bugs.python.org/issue31014 - URLs might not be opened in the correct browser")
+            webbrowser.open(url)
 
 
 def goto(url_id):
